@@ -11,7 +11,9 @@ module Data.HashTable.Test.Common
   ) where
 
 ------------------------------------------------------------------------------
+import           Control.Applicative                  (pure, (<|>))
 import           Control.Monad                        (foldM_, liftM, when)
+import qualified Control.Monad                        as Monad
 import           Data.IORef
 import           Data.List                            hiding (delete, insert,
                                                        lookup)
@@ -62,6 +64,12 @@ announce _ _ = return ()
 #endif
 
 
+------------------------------------------------------------------------------
+announceQ :: Show a => String -> a -> PropertyM IO ()
+announceQ nm x = run $ announce nm x
+
+
+------------------------------------------------------------------------------
 assertEq :: (Eq a, Show a) =>
             String -> a -> a -> PropertyM IO ()
 assertEq s expected got =
@@ -95,6 +103,7 @@ tests prefix dummyArg = testGroup prefix $ map f ts
          , SomeTest testDelete
          , SomeTest testNastyFullLookup
          , SomeTest testForwardSearch3
+         , SomeTest testMutate
          ]
 
 
@@ -109,8 +118,8 @@ testFromListToList prefix dummyArg =
   where
     prop :: GenIO -> [(Int, Int)] -> PropertyM IO ()
     prop rng origL = do
-        let l = V.toList $ shuffle rng $ V.fromList $ dedupe origL
-        run $ announce "fromListToList" l
+        let l = V.toList $ shuffleVector rng $ V.fromList $ dedupe origL
+        announceQ "fromListToList" l
         ht <- run $ fromList l
         l' <- run $ toList ht
         assertEq "fromList . toList == id" (sort l) (sort l')
@@ -128,8 +137,8 @@ testInsert prefix dummyArg =
   where
     prop :: GenIO -> ([(Int, Int)], (Int,Int)) -> PropertyM IO ()
     prop rng o@(origL, (k,v)) = do
-        run $ announce "insert" o
-        let l = V.toList $ shuffle rng $ V.fromList $ remove k $ dedupe origL
+        announceQ "insert" o
+        let l = V.toList $ shuffleVector rng $ V.fromList $ remove k $ dedupe origL
         assert $ all (\t -> fst t /= k) l
 
         ht <- run $ fromList l
@@ -154,8 +163,8 @@ testInsert2 prefix dummyArg =
   where
     prop :: GenIO -> ([(Int, Int)], (Int,Int,Int)) -> PropertyM IO ()
     prop rng o@(origL, (k,v,v2)) = do
-        run $ announce "insert2" o
-        let l = V.toList $ shuffle rng $ V.fromList $ dedupe origL
+        announceQ "insert2" o
+        let l = V.toList $ shuffleVector rng $ V.fromList $ dedupe origL
         ht   <- run $ fromList l
 
         run $ insert ht k v
@@ -178,7 +187,7 @@ testNewAndInsert prefix dummyArg =
   where
     prop :: (Int,Int,Int) -> PropertyM IO ()
     prop o@(k,v,v2) = do
-        run $ announce "newAndInsert" o
+        announceQ "newAndInsert" o
         ht <- run new
 
         nothing <- run $ lookup ht k
@@ -225,7 +234,7 @@ testGrowTable prefix dummyArg =
 
     prop :: Int -> PropertyM IO ()
     prop n = do
-        run $ announce "growTable" n
+        announceQ "growTable" n
         ht <- run $ go n
         i <- liftM head $ run $ sample' $ choose (0,n-1)
 
@@ -271,7 +280,7 @@ testDelete prefix dummyArg =
 
     prop :: Int -> PropertyM IO ()
     prop n = do
-        run $ announce "delete" n
+        announceQ "delete" n
 
         ht <- run $ go n
 
@@ -287,6 +296,34 @@ testDelete prefix dummyArg =
         forceType dummyArg ht
 
 
+------------------------------------------------------------------------------
+testMutate :: HashTest
+testMutate prefix dummyArg = testProperty (prefix ++ "/mutate") $
+                             monadicIO $ forAllM arbitrary prop
+  where
+    prop :: ([(Int, Int)],  [(Int, [Int])]) -> PropertyM IO ()
+    prop o@(seedList, testList) = do
+      announceQ "mutate" o
+      ht <- run $ fromList seedList
+      Monad.mapM_ (testOne ht) testList
+
+
+    upd n v = (fmap (+ n) v <|> pure n, ())
+
+    testOne ht (k, values) = do
+        pre . not . null $ values
+        run $ mutate ht k (const (Nothing, ()))
+        out1 <- run $ lookup ht k
+        assertEq ("mutate deletes " ++ show k) Nothing out1
+        out2 <- run $ do
+            Monad.mapM_ (mutate ht k . upd) values
+            (Just v) <- lookup ht k
+            return $! v
+        let s = sum values
+        assertEq "mutate inserts correctly folded list value" s out2
+        forceType dummyArg ht
+
+
 ------------------------------------------------------------------------------
 data Action = Lookup Int
             | Insert Int
@@ -477,8 +514,8 @@ remove m l = go id l
 
 
 ------------------------------------------------------------------------------
-shuffle :: GenIO -> Vector k -> Vector k
-shuffle rng v = if V.null v then v else V.modify go v
+shuffleVector :: GenIO -> Vector k -> Vector k
+shuffleVector rng v = if V.null v then v else V.modify go v
   where
     !n = V.length v
 
